# Import the SDK and required libraries
import boto3
import json
import os
import logging
import time
import calendar
from datetime import datetime



from botocore.exceptions import ClientError

logger = logging.getLogger()
logger.setLevel(logging.INFO)



# Getting the parameters from LAMBDA Function Environment Variable
namespace = os.environ['NAMESPACE']
sns_arn = os.environ['SNS_TOPIC']
s3bucketforuploads = os.environ['S3Bucket_For_Message_Uploads']
AppLogsEnabled = os.environ['ApplicationLogsEnabled']
CWLogGroupName = os.environ['CWLogGroupName']
elb_name_tag = os.environ['LOAD_BALANCER_NAME_TAG']

if namespace == 'AWS/ELB':
	tg_arn = None
	tg_type = None
else:
	tg_arn = os.environ['TARGETGROUP_ARN'].strip()
	tg_type = (os.environ['TARGETGROUP_TYPE'].strip()).lower()



def lambda_handler(event, context):
	"""
	Main Lambda handler
	"""
	# print("Received event: " + json.dumps(event, indent=2))
	print(event)
	message = event['Records'][0]['Sns']['Message']
	global sns_client
	global elb_client
	global elbv2_client
	global ec2_client
	global cw_logs_client
	global s3_client

	try:
		sns_client = boto3.client('sns')
	except ClientError as e:
		logger.error(e.response['Error']['Message'])

	try:
		s3_client = boto3.client('s3')
	except ClientError as e:
		logger.error(e.response['Error']['Message'])

	try:
		elb_client = boto3.client('elb')
	except ClientError as e:
		logger.error(e.response['Error']['Message'])

	try:
		elbv2_client = boto3.client('elbv2')
	except ClientError as e:
		logger.error(e.response['Error']['Message'])

	try:
		ec2_client = boto3.client('ec2')
	except ClientError as e:
		logger.error(e.response['Error']['Message'])

	try:
		cw_logs_client = boto3.client('logs')
	except ClientError as e:
		logger.error(e.response['Error']['Message'])


	if "AlarmName" in message:
		json_message = json.loads(message)
		accountid = str(json_message['AWSAccountId'])
		alarm_trigger = str(json_message['NewStateValue'])
		timestamp = str(json_message['StateChangeTime'])
		originating_alarmname = str(json_message['AlarmName'])
		if "LOW" in originating_alarmname:
			severity = "LOW"
		elif "MED" in originating_alarmname:
			severity = "MED"
		elif "HIGH" in originating_alarmname:
			severity = "HIGH"
		else:
			severity = "UNKNOWN"
		region = os.environ["AWS_REGION"]
		logger.info("=======Start Lambda Function=======")
		logger.info("AccountID:{}".format(accountid))
		logger.info("Region:{}".format(region))
		logger.info("Namespace:{}".format(namespace))
		logger.info("Alarm State:{}".format(alarm_trigger))
		if namespace == "AWS/ApplicationELB":
			elb_name = os.environ['LOAD_BALANCER_NAME']
			print("Printing the ALB for check")
			print(elb_name)
		elif namespace == "AWS/NetworkELB":
			elb_name = os.environ['LOAD_BALANCER_NAME']
			print("Printing the NLB Name for check")
			print(elb_name)

		elif namespace == 'AWS/ELB':
			elb_name = os.environ['LOAD_BALANCER_NAME']
			print("Printing the CLB Name for check")
			print(elb_name)

		# Take actions when an Alarm is triggered
		if alarm_trigger == 'ALARM':
			if namespace == "AWS/ApplicationELB":
				lbtype='elbv2/alb'
				#Get health check port from describe target health as one target group can have different health check ports for each target
				unhealthy_list,unhealthy_targets_ip_port_list = describe_target_health(tg_arn, elb_name, tg_type)
				number_of_instances = len(unhealthy_list)
				print(number_of_instances)
				if number_of_instances > 0:
					alarmname = "{}".format(originating_alarmname)
					if AppLogsEnabled == "TRUE":
						endtime=calendar.timegm(time.gmtime())
						starttime= endtime - 600
						starttime_ms = starttime*1000
						endtime_ms = endtime*1000
						messages_list = []
						print("list of servers marked unhealthy")
						print(unhealthy_list)
						for target_instance in unhealthy_list:
							if tg_type == 'instance':
								instance_id = target_instance['Target']['Id']
								print(instance_id)
								try:
									CW_Logs_Events = cw_logs_client.get_log_events(logGroupName=CWLogGroupName,logStreamName=instance_id,startTime=starttime_ms,endTime=endtime_ms,limit=10)
									messages_list.append("These are the Application Logs for the Server with InstanceId: "+instance_id)
									for message in CW_Logs_Events['events']:
										message_time=message['timestamp']/1000
										message_time_standardformat=time.strftime('%Y-%m-%d %H:%M:%S:%s', time.localtime(message_time))
										messages_list.append(message_time_standardformat+"-----"+message['message'])
								except ClientError as e:
									error_message = e.response['Error']['Message']
									messages_list.append("Error Occurred while getting logs for the instance - "+instance_id)
									messages_list.append("Error Message: "+error_message)
									messages_list.append("Please validate if the log group: "+CWLogGroupName+" and logstream: "+instance_id+" exists in the account and region")
								
							elif tg_type == 'ip':
								instance_id = target_instance['Target']['Id']
								instance_details_from_ip = ec2_client.describe_instances(Filters=[{'Name':'network-interface.addresses.private-ip-address','Values':[instance_id]}])
								if len(instance_details_from_ip['Reservations']) != 0:
									instance_id_from_ip = instance_details_from_ip['Reservations'][0]['Instances'][0]['InstanceId']
									print(instance_id_from_ip)
									try:
										CW_Logs_Events = cw_logs_client.get_log_events(logGroupName=CWLogGroupName,logStreamName=instance_id_from_ip,startTime=starttime_ms,endTime=endtime_ms,limit=10)
										messages_list.append("These are the Application Logs for the IP Based Target: "+instance_id+" with the InstanceId: "+instance_id_from_ip)
										for message in CW_Logs_Events['events']:
											message_time=message['timestamp']/1000
											message_time_standardformat=time.strftime('%Y-%m-%d %H:%M:%S:%s', time.localtime(message_time))
											messages_list.append(message_time_standardformat+"-----"+message['message'])
									except ClientError as e:
										error_message = e.response['Error']['Message']
										messages_list.append("Error Occurred while getting logs for the instance - "+instance_id_from_ip)
										messages_list.append("Error Message: "+error_message)
										messages_list.append("Please validate if the log group: "+CWLogGroupName+" and logstream: "+instance_id_from_ip+" exists in the account and region")


								else:
									messages_list.append("IP Based Target:{}  , is not part of this Account. NO LOGS TO DISPLAY".format(instance_id))
					else:
						messages_list = []
						messages_list.append("APPLICATION LOGS ARE NOT ENABLED. NO LOGS TO DISPLAY")

					send_sns(alarmname, number_of_instances, unhealthy_list, elb_name, region, timestamp, accountid, lbtype, tg_arn, severity, messages_list)
				else:
					print("No need to take action as there are no unhealthy targets")
			elif namespace == "AWS/NetworkELB":
				lbtype='elbv2/nlb'
				#Get health check port from describe target health as one target group can have different health check ports for each target
				unhealthy_list,unhealthy_targets_ip_port_list = describe_target_health(tg_arn, elb_name, tg_type)
				number_of_instances = len(unhealthy_list)
				print(number_of_instances)
				if number_of_instances > 0:
					alarmname = "{}".format(originating_alarmname)
					if AppLogsEnabled == "TRUE":
						endtime=calendar.timegm(time.gmtime())
						starttime= endtime - 600
						starttime_ms = starttime*1000
						endtime_ms = endtime*1000
						messages_list = []
						print("list of servers marked unhealthy")
						print(unhealthy_list)
						for target_instance in unhealthy_list:
							if tg_type == 'instance':
								instance_id = target_instance['Target']['Id']
								print(instance_id)
								try:
									CW_Logs_Events = cw_logs_client.get_log_events(logGroupName=CWLogGroupName,logStreamName=instance_id,startTime=starttime_ms,endTime=endtime_ms,limit=10)
									messages_list.append("These are the Application Logs for the Server with InstanceId: "+instance_id)
									for message in CW_Logs_Events['events']:
										message_time=message['timestamp']/1000
										message_time_standardformat=time.strftime('%Y-%m-%d %H:%M:%S:%s', time.localtime(message_time))
										messages_list.append(message_time_standardformat+"-----"+message['message'])
								except ClientError as e:
									error_message = e.response['Error']['Message']
									messages_list.append("Error Occurred while getting logs for the instance - "+instance_id)
									messages_list.append("Error Message: "+error_message)
									messages_list.append("Please validate if the log group: "+CWLogGroupName+" and logstream: "+instance_id+" exists in the account and region")

							elif tg_type == 'ip':
								instance_id = target_instance['Target']['Id']
								instance_details_from_ip = ec2_client.describe_instances(Filters=[{'Name':'network-interface.addresses.private-ip-address','Values':[instance_id]}])
								if len(instance_details_from_ip['Reservations']) != 0:
									instance_id_from_ip = instance_details_from_ip['Reservations'][0]['Instances'][0]['InstanceId']
									print(instance_id_from_ip)
									try:
										CW_Logs_Events = cw_logs_client.get_log_events(logGroupName=CWLogGroupName,logStreamName=instance_id_from_ip,startTime=starttime_ms,endTime=endtime_ms,limit=10)
										messages_list.append("These are the Application Logs for the IP Based Target: "+instance_id+" with the InstanceId: "+instance_id_from_ip)
										for message in CW_Logs_Events['events']:
											message_time=message['timestamp']/1000
											message_time_standardformat=time.strftime('%Y-%m-%d %H:%M:%S:%s', time.localtime(message_time))
											messages_list.append(message_time_standardformat+"-----"+message['message'])
									except ClientError as e:
										error_message = e.response['Error']['Message']
										messages_list.append("Error Occurred while getting logs for the instance - "+instance_id_from_ip)
										messages_list.append("Error Message: "+error_message)
										messages_list.append("Please validate if the log group: "+CWLogGroupName+" and logstream: "+instance_id_from_ip+" exists in the account and region")

								else:
									messages_list.append("IP Based Target:{}  , is not part of this Account. NO LOGS TO DISPLAY".format(instance_id))

					else:
						messages_list = []
						messages_list.append("APPLICATION LOGS ARE NOT ENABLED. NO LOGS TO DISPLAY")

					send_sns(alarmname, number_of_instances, unhealthy_list, elb_name, region, timestamp, accountid, lbtype, tg_arn, severity, messages_list)
				else:
					print("No need to take action as there are no unhealthy targets")

			elif namespace == "AWS/ELB":
				lbtype='elb'
				# Get health check port from healthcheck configuration for CLB
				unhealthy_list,unhealthy_instances_id_list = describe_instance_health(elb_name)
				number_of_instances = len(unhealthy_instances_id_list)
				print(number_of_instances)
				if number_of_instances > 0:
					alarmname = "{}".format(originating_alarmname)
					if AppLogsEnabled == "TRUE":
						endtime=calendar.timegm(time.gmtime())
						starttime= endtime - 600
						starttime_ms = starttime*1000
						endtime_ms = endtime*1000
						messages_list = []
						print("list of servers marked unhealthy")
						print(unhealthy_list)
						for target_instance in unhealthy_list:
							instance_id = target_instance['InstanceId']
							print(instance_id)
							try:
								CW_Logs_Events = cw_logs_client.get_log_events(logGroupName=CWLogGroupName,logStreamName=instance_id,startTime=starttime_ms,endTime=endtime_ms,limit=10)
								messages_list.append("These are the Application Logs for the Server with InstanceId: "+instance_id)
								for message in CW_Logs_Events['events']:
									message_time=message['timestamp']/1000
									message_time_standardformat=time.strftime('%Y-%m-%d %H:%M:%S:%s', time.localtime(message_time))
									messages_list.append(message_time_standardformat+"-----"+message['message'])
							except ClientError as e:
									error_message = e.response['Error']['Message']
									messages_list.append("Error Occurred while getting logs for the instance - "+instance_id)
									messages_list.append("Error Message: "+error_message)
									messages_list.append("Please validate if the log group: "+CWLogGroupName+" and logstream: "+instance_id+" exists in the account and region")

					else:
						messages_list = []
						messages_list.append("APPLICATION LOGS ARE NOT ENABLED. NO LOGS TO DISPLAY")

					send_sns(alarmname, number_of_instances, unhealthy_list, elb_name, region, timestamp, accountid, lbtype, tg_arn, severity, messages_list)
				else:
					print("No need to take action as there are no unhealthy targets")


def describe_instance_health(elb_name):
	"""
	Describe targets health of ClassicLoadBalancer
	"""
	unhealthy_list = []
	unhealthy_instances_id_list = []
	try:
		health_response = elb_client.describe_instance_health(LoadBalancerName=elb_name)
	except ClientError as e:
		logger.error(e.response['Error']['Message'])
	for instance in health_response["InstanceStates"]:
		if instance["State"] == "OutOfService":
			unhealthy_list.append(instance)
			unhealthy_instances_id_list.append(instance['InstanceId'])

	logger.info("elb - Unhealthy instances: {}".format(unhealthy_list))
	return unhealthy_list,unhealthy_instances_id_list


def send_sns(alarmname, number_of_instances, unhealthy_list, elb_name, region, timestamp, accountid,lbtype, tg_arn, severity, messages_list):
	"""
	Send notification to SNS topic subscribers of unhealthy instances/targets
	"""
	logger.info("\n==SNS notification meta data==\n")
	logger.info("timestamp:{}\n accountid:{}\n region:{}\n elb_name:{}\n unhealthy_list:{}\n".format(timestamp, accountid, region, elb_name, unhealthy_list))
	unhealthy_message_list = []
	for target_instance in unhealthy_list:
		if lbtype == 'elbv2/alb' or lbtype =='elbv2/nlb' :
			instance_id = target_instance['Target']['Id']
			print(instance_id)
			instance_healthcheck_port = target_instance['HealthCheckPort']
			instance_healthcheck_state = target_instance['TargetHealth']['State']
			instance_healthcheck_description = target_instance['TargetHealth']['Description']
			instance_healthcheck_reason = target_instance['TargetHealth']['Reason']
			if tg_type == 'ip':
				instance_details_from_ip = ec2_client.describe_instances(Filters=[{'Name':'network-interface.addresses.private-ip-address','Values':[instance_id]}])
				if len(instance_details_from_ip['Reservations']) != 0:
					instance_id_from_ip = instance_details_from_ip['Reservations'][0]['Instances'][0]['InstanceId']
					instance_tags = instance_details_from_ip['Reservations'][0]['Instances'][0]['Tags']
					for ec2_tag in instance_tags:
						if ec2_tag['Key'] == 'Name':
							ec2_name = ec2_tag['Value']
					unhealthy_status_message = "Instance - (IP:{}) (InstanceId:{}) (InstanceName:{}) is marked unhealthy by the (Load Balancer)-{} on port {}. Reason: {} - {}".format(instance_id,instance_id_from_ip,ec2_name,elb_name,instance_healthcheck_port,instance_healthcheck_description,instance_healthcheck_reason)
					unhealthy_message_list.append(unhealthy_status_message)
				else:
					unhealthy_status_message = "Instance - (IP:{}) (InstanceId: NOT AVAIALBLE IN THIS ACCOUNT ) (InstanceName: NOT AVAIALBLE IN THIS ACCOUNT) is marked unhealthy by the (Load Balancer)-{} on port {}. Reason: {} - {}".format(instance_id,elb_name,instance_healthcheck_port,instance_healthcheck_description,instance_healthcheck_reason)
					unhealthy_message_list.append(unhealthy_status_message)

			elif tg_type == 'instance':
				instance_info = ec2_client.describe_instances(InstanceIds=[instance_id])['Reservations'][0]['Instances'][0]
				instance_tags = instance_info['Tags']
				for ec2_tag in instance_tags:
					if ec2_tag['Key'] == 'Name':
						ec2_name = ec2_tag['Value']
				unhealthy_status_message = "Instance - (InstanceId:{}) (InstanceName:{}) is marked unhealthy by the (Load Balancer)-{} on port {}. Reason: {} - {}".format(instance_id,ec2_name,elb_name,instance_healthcheck_port,instance_healthcheck_description,instance_healthcheck_reason)
				unhealthy_message_list.append(unhealthy_status_message)

		elif lbtype=='elb':
			instance_id = target_instance['InstanceId']
			print(instance_id)
			
			instance_healthcheck_state = target_instance['State']
			instance_healthcheck_description = target_instance['Description']
			instance_healthcheck_reason = target_instance['ReasonCode']
			response_healthcheck=elb_client.describe_load_balancers(LoadBalancerNames=[elb_name])
			instance_healthcheck_port = response_healthcheck['LoadBalancerDescriptions'][0]['HealthCheck']['Target'].split(":")[1].split("/")[0]
			instance_info = ec2_client.describe_instances(InstanceIds=[instance_id])['Reservations'][0]['Instances'][0]
			instance_tags = instance_info['Tags']
			for ec2_tag in instance_tags:
					if ec2_tag['Key'] == 'Name':
						ec2_name = ec2_tag['Value']

			unhealthy_status_message = "Instance - (InstanceId:{}) (InstanceName:{}) is marked unhealthy by the (Load Balancer)-{} on port {}. Reason: {} - {}".format(instance_id,ec2_name,elb_name,instance_healthcheck_port,instance_healthcheck_description,instance_healthcheck_reason)
			unhealthy_message_list.append(unhealthy_status_message)


	print("Sending SNS Notification test with custom message")
	print(unhealthy_message_list)
	referecelink = "Please refer these documents to troubleshoot the reason for failed health checks - https://aws.amazon.com/premiumsupport/knowledge-center/elb-fix-failing-health-checks-alb/ ; https://docs.aws.amazon.com/elasticloadbalancing/latest/application/target-group-health-checks.html#target-health-reason-codes"
	
	if lbtype == 'elbv2/alb' or lbtype =='elbv2/nlb' :
		
		message = "AlarmName: {} \nNumberOfTargets_Failed_HealthChecks: {} \nAWSAccountId: {} \nTimestamp: {} \nRegion: {} \nELB: {} \nTargetGroupARN: {}" \
			          "\n\n\nUnhealthy Targets and Cause of Failure: \n{}" \
			          "\n\n\n====ApplicationLogs for last 10 minutes (LIMIT OF 10 Events for rest of the events please check the CW log group: {})=====: \n{}" \
			          "\n\n\n====For more information about how to troubleshoot unhealthy targets, please refer to the following links====: \n{}" \
				.format(alarmname, number_of_instances, accountid, timestamp, region, elb_name, tg_arn, "\n".join(unhealthy_message_list), CWLogGroupName, "\n".join(messages_list), referecelink)

	elif lbtype == 'elb':
		
		message = "AlarmName: {} \nNumberOfInstances_Failed_HealthChecks: {} \nAWSAccountId: {} \nTimestamp: {} \nRegion: {} \nELB: {} " \
			          "\n\n\nUnhealthy Instances and Cause of Failure: \n{}" \
			          "\n\n\n====ApplicationLogs for last 10 minutes (LIMIT OF 10 Events for rest of the events please check the CW log group: {})=====: \n{}" \
			          "\n\n\n====For more information about how to troubleshoot unhealthy targets, please refer to the following links====: \n{}" \
				.format(alarmname, number_of_instances, accountid, timestamp, region, elb_name, "\n".join(unhealthy_message_list), CWLogGroupName, "\n".join(messages_list), referecelink)

	try:
		sns_client.publish(
			TopicArn=sns_arn,
			Message=message,
			Subject='Alarm: ELB unhealthy targets - ' + elb_name_tag,
			MessageStructure='string'
		)
	except ClientError as e:
		logger.error(e.response['Error']['Message'])

	#copying the message to the S3 bucket as a backup
	currenttime= datetime.now()
	filename="UnhealthyTargetsNotification_"+elb_name+str(currenttime)+".txt"
	filebody=bytes(message, 'utf-8')
	s3_client.put_object(Bucket=s3bucketforuploads,Body=filebody,Key=filename)


def describe_target_health(targetgroup_arn, elb_name, tg_type):
	"""
	Describe targets health of Application/NetworkLoadBalancer
	"""
	unhealthy_list = []
	unhealthy_targets_id_port_list=[]
	unhealthy_targets_ip_port_list = []
	try:
		response = elbv2_client.describe_target_health(TargetGroupArn=targetgroup_arn)
	except ClientError as e:
		logger.error(e.response['Error']['Message'])
	for target in response['TargetHealthDescriptions']:
		if target['TargetHealth']['State'] == 'unhealthy':
			unhealthy_list.append(target)
			unhealthy_targets_id_port_list.append(target['Target']['Id'] + ":" + str(target['Target']['Port']))
	if tg_type == 'ip':
		unhealthy_targets_ip_port_list=unhealthy_targets_id_port_list
	if tg_type == 'instance':
		unhealthy_targets_ip_port_list = get_ondemand_target_port(unhealthy_targets_id_port_list)
	logger.info("elbv2 - Unhealthy targets: {}".format(unhealthy_list))
	logger.info("elbv2 - unhealthy_targets_ip_port_list: {}".format(unhealthy_targets_ip_port_list))
	return unhealthy_list,unhealthy_targets_ip_port_list


def get_ondemand_target_port(unhealthy_instances_dimension_list):
	target_id_list=[]
	target_port_list=[]
	targets_ip_port_list = []
	for target in unhealthy_instances_dimension_list:
		target_id = target.split(':')[0]
		target_port = target.split(':')[1]
		target_id_list.append(target_id)
		target_port_list.append(target_port)
	target_ip_list = get_instance_private_ip(target_id_list)
	for ip,port in zip(target_ip_list,target_port_list):
		targets_ip_port_list.append("{}:{}".format(ip,port))
	return targets_ip_port_list


def get_instance_private_ip(unhealthy_instance_id_list):
	unhealthy_instance_ip_list = []
	response = ec2_client.describe_instances(InstanceIds=unhealthy_instance_id_list)

	for group in response['Reservations']:
		for instance in group['Instances']:
			for eni in instance['NetworkInterfaces']:
				if eni['Attachment']['DeviceIndex'] == 0:
					for private_ip in eni['PrivateIpAddresses']:
						if private_ip['Primary'] == True:
							unhealthy_instance_ip_list.append(private_ip['PrivateIpAddress'])
	return unhealthy_instance_ip_list